@@ -1,14 +1,13 @@
-import importlib
 import inspect
-import times
 from uuid import uuid4
 try:
     from cPickle import loads, dumps, UnpicklingError
-except ImportError: # noqa
+except ImportError:  # noqa
     from pickle import loads, dumps, UnpicklingError  # noqa
 from .local import LocalStack
 from .connections import resolve_connection
 from .exceptions import UnpickleError, NoSuchJobError
+from .utils import import_attribute, utcnow, utcformat, utcparse
 from rq.compat import text_type, decode_redis_hash, as_text
 
 
@@ -16,8 +15,13 @@ def enum(name, *sequential, **named):
     values = dict(zip(sequential, range(len(sequential))), **named)
     return type(name, (), values)
 
-Status = enum('Status', QUEUED='queued', FINISHED='finished', FAILED='failed',
-                        STARTED='started')
+Status = enum('Status',
+              QUEUED='queued', FINISHED='finished', FAILED='failed',
+              STARTED='started')
+
+# Sentinel value to mark that some of our lazily evaluated properties have not
+# yet been evaluated.
+UNEVALUATED = object()
 
 
 def unpickle(pickled_string):
@@ -52,23 +56,24 @@ def requeue_job(job_id, connection=None):
     fq.requeue(job_id)
 
 
-def get_current_job():
+def get_current_job(connection=None):
     """Returns the Job instance that is currently being executed.  If this
     function is invoked from outside a job context, None is returned.
     """
     job_id = _job_stack.top
     if job_id is None:
         return None
-    return Job.fetch(job_id)
+    return Job.fetch(job_id, connection=connection)
 
 
 class Job(object):
     """A Job is just a convenient datastructure to pass around job (meta) data.
     """
+
     # Job construction
     @classmethod
     def create(cls, func, args=None, kwargs=None, connection=None,
-               result_ttl=None, status=None):
+               result_ttl=None, status=None, description=None, depends_on=None, timeout=None):
         """Creates a new Job instance for the given function, arguments, and
         keyword arguments.
         """
@@ -76,9 +81,16 @@ class Job(object):
             args = ()
         if kwargs is None:
             kwargs = {}
-        assert isinstance(args, (tuple, list)), '%r is not a valid args list.' % (args,)
-        assert isinstance(kwargs, dict), '%r is not a valid kwargs dict.' % (kwargs,)
+
+        if not isinstance(args, (tuple, list)):
+            raise TypeError('{0!r} is not a valid args list.'.format(args))
+        if not isinstance(kwargs, dict):
+            raise TypeError('{0!r} is not a valid kwargs dict.'.format(kwargs))
+
         job = cls(connection=connection)
+
+        # Set the core job tuple properties
+        job._instance = None
         if inspect.ismethod(func):
             job._instance = func.__self__
             job._func_name = func.__name__
@@ -88,40 +100,69 @@ class Job(object):
             job._func_name = func
         job._args = args
         job._kwargs = kwargs
-        job.description = job.get_call_string()
+
+        # Extra meta data
+        job.description = description or job.get_call_string()
         job.result_ttl = result_ttl
+        job.timeout = timeout
         job._status = status
+
+        # dependency could be job instance or id
+        if depends_on is not None:
+            job._dependency_id = depends_on.id if isinstance(depends_on, Job) else depends_on
         return job
 
-    @property
-    def func_name(self):
-        return self._func_name
-
-    def _get_status(self):
+    def get_status(self):
         self._status = as_text(self.connection.hget(self.key, 'status'))
         return self._status
 
-    def _set_status(self, status):
+    def _get_status(self):
+        raise DeprecationWarning(
+            "job.status is deprecated. Use job.get_status() instead"
+        )
+        return self.get_status()
+
+    def set_status(self, status):
         self._status = status
         self.connection.hset(self.key, 'status', self._status)
 
+    def _set_status(self, status):
+        raise DeprecationWarning(
+            "job.status is deprecated. Use job.set_status() instead"
+        )
+        self.set_status(status)
+
     status = property(_get_status, _set_status)
 
     @property
     def is_finished(self):
-        return self.status == Status.FINISHED
+        return self.get_status() == Status.FINISHED
 
     @property
     def is_queued(self):
-        return self.status == Status.QUEUED
+        return self.get_status() == Status.QUEUED
 
     @property
     def is_failed(self):
-        return self.status == Status.FAILED
+        return self.get_status() == Status.FAILED
 
     @property
     def is_started(self):
-        return self.status == Status.STARTED
+        return self.get_status() == Status.STARTED
+
+    @property
+    def dependency(self):
+        """Returns a job's dependency. To avoid repeated Redis fetches, we cache
+        job.dependency as job._dependency.
+        """
+        if self._dependency_id is None:
+            return None
+        if hasattr(self, '_dependency'):
+            return self._dependency
+        job = Job.fetch(self._dependency_id, connection=self.connection)
+        job.refresh()
+        self._dependency = job
+        return job
 
     @property
     def func(self):
@@ -132,22 +173,82 @@ class Job(object):
         if self.instance:
             return getattr(self.instance, func_name)
 
-        module_name, func_name = func_name.rsplit('.', 1)
-        module = importlib.import_module(module_name)
-        return getattr(module, func_name)
+        return import_attribute(self.func_name)
+
+    def _unpickle_data(self):
+        self._func_name, self._instance, self._args, self._kwargs = unpickle(self.data)
+
+    @property
+    def data(self):
+        if self._data is UNEVALUATED:
+            if self._func_name is UNEVALUATED:
+                raise ValueError('Cannot build the job data.')
+
+            if self._instance is UNEVALUATED:
+                self._instance = None
+
+            if self._args is UNEVALUATED:
+                self._args = ()
+
+            if self._kwargs is UNEVALUATED:
+                self._kwargs = {}
+
+            job_tuple = self._func_name, self._instance, self._args, self._kwargs
+            self._data = dumps(job_tuple)
+        return self._data
+
+    @data.setter
+    def data(self, value):
+        self._data = value
+        self._func_name = UNEVALUATED
+        self._instance = UNEVALUATED
+        self._args = UNEVALUATED
+        self._kwargs = UNEVALUATED
+
+    @property
+    def func_name(self):
+        if self._func_name is UNEVALUATED:
+            self._unpickle_data()
+        return self._func_name
+
+    @func_name.setter
+    def func_name(self, value):
+        self._func_name = value
+        self._data = UNEVALUATED
 
     @property
     def instance(self):
+        if self._instance is UNEVALUATED:
+            self._unpickle_data()
         return self._instance
 
+    @instance.setter
+    def instance(self, value):
+        self._instance = value
+        self._data = UNEVALUATED
+
     @property
     def args(self):
+        if self._args is UNEVALUATED:
+            self._unpickle_data()
         return self._args
 
+    @args.setter
+    def args(self, value):
+        self._args = value
+        self._data = UNEVALUATED
+
     @property
     def kwargs(self):
+        if self._kwargs is UNEVALUATED:
+            self._unpickle_data()
         return self._kwargs
 
+    @kwargs.setter
+    def kwargs(self, value):
+        self._kwargs = value
+        self._data = UNEVALUATED
+
     @classmethod
     def exists(cls, job_id, connection=None):
         """Returns whether a job hash exists for the given job ID."""
@@ -163,23 +264,15 @@ class Job(object):
         job.refresh()
         return job
 
-    @classmethod
-    def safe_fetch(cls, id, connection=None):
-        """Fetches a persisted job from its corresponding Redis key, but does
-        not instantiate it, making it impossible to get UnpickleErrors.
-        """
-        job = cls(id, connection=connection)
-        job.refresh(safe=True)
-        return job
-
     def __init__(self, id=None, connection=None):
         self.connection = resolve_connection(connection)
         self._id = id
-        self.created_at = times.now()
-        self._func_name = None
-        self._instance = None
-        self._args = None
-        self._kwargs = None
+        self.created_at = utcnow()
+        self._data = UNEVALUATED
+        self._func_name = UNEVALUATED
+        self._instance = UNEVALUATED
+        self._args = UNEVALUATED
+        self._kwargs = UNEVALUATED
         self.description = None
         self.origin = None
         self.enqueued_at = None
@@ -189,8 +282,11 @@ class Job(object):
         self.timeout = None
         self.result_ttl = None
         self._status = None
+        self._dependency_id = None
         self.meta = {}
 
+    def __repr__(self):  # noqa
+        return 'Job(%r, enqueued_at=%r)' % (self._id, self.enqueued_at)
 
     # Data access
     def get_id(self):  # noqa
@@ -212,16 +308,20 @@ class Job(object):
         """The Redis key that is used to store job hash under."""
         return b'rq:job:' + job_id.encode('utf-8')
 
+    @classmethod
+    def dependents_key_for(cls, job_id):
+        """The Redis key that is used to store job hash under."""
+        return 'rq:job:%s:dependents' % (job_id,)
+
     @property
     def key(self):
         """The Redis key that is used to store job hash under."""
         return self.key_for(self.id)
 
-    @property  # noqa
-    def job_tuple(self):
-        """Returns the job tuple that encodes the actual function call that
-        this job represents."""
-        return (self.func_name, self.instance, self.args, self.kwargs)
+    @property
+    def dependents_key(self):
+        """The Redis key that is used to store job hash under."""
+        return self.dependents_key_for(self.id)
 
     @property
     def result(self):
@@ -250,9 +350,8 @@ class Job(object):
     """Backwards-compatibility accessor property `return_value`."""
     return_value = result
 
-
     # Persistence
-    def refresh(self, safe=False):  # noqa
+    def refresh(self):  # noqa
         """Overwrite the current instance's properties with the values in the
         corresponding Redis key.
 
@@ -265,20 +364,15 @@ class Job(object):
 
         def to_date(date_str):
             if date_str is None:
-                return None
+                return
             else:
-                return times.to_universal(as_text(date_str))
+                return utcparse(as_text(date_str))
 
         try:
             self.data = obj['data']
         except KeyError:
             raise NoSuchJobError('Unexpected job format: {0}'.format(obj))
 
-        try:
-            self._func_name, self._instance, self._args, self._kwargs = unpickle(self.data)
-        except UnpickleError:
-            if not safe:
-                raise
         self.created_at = to_date(as_text(obj.get('created_at')))
         self.origin = as_text(obj.get('origin'))
         self.description = as_text(obj.get('description'))
@@ -287,28 +381,25 @@ class Job(object):
         self._result = unpickle(obj.get('result')) if obj.get('result') else None  # noqa
         self.exc_info = obj.get('exc_info')
         self.timeout = int(obj.get('timeout')) if obj.get('timeout') else None
-        self.result_ttl = int(obj.get('result_ttl')) if obj.get('result_ttl') else None # noqa
+        self.result_ttl = int(obj.get('result_ttl')) if obj.get('result_ttl') else None  # noqa
         self._status = as_text(obj.get('status') if obj.get('status') else None)
+        self._dependency_id = as_text(obj.get('dependency_id', None))
         self.meta = unpickle(obj.get('meta')) if obj.get('meta') else {}
 
-    def save(self, pipeline=None):
-        """Persists the current job instance to its corresponding Redis key."""
-        key = self.key
-        connection = pipeline if pipeline is not None else self.connection
-
+    def dump(self):
+        """Returns a serialization of the current job instance"""
         obj = {}
-        obj['created_at'] = times.format(self.created_at or times.now(), 'UTC')
+        obj['created_at'] = utcformat(self.created_at or utcnow())
+        obj['data'] = self.data
 
-        if self.func_name is not None:
-            obj['data'] = dumps(self.job_tuple)
         if self.origin is not None:
             obj['origin'] = self.origin
         if self.description is not None:
             obj['description'] = self.description
         if self.enqueued_at is not None:
-            obj['enqueued_at'] = times.format(self.enqueued_at, 'UTC')
+            obj['enqueued_at'] = utcformat(self.enqueued_at)
         if self.ended_at is not None:
-            obj['ended_at'] = times.format(self.ended_at, 'UTC')
+            obj['ended_at'] = utcformat(self.ended_at)
         if self._result is not None:
             obj['result'] = dumps(self._result)
         if self.exc_info is not None:
@@ -319,10 +410,19 @@ class Job(object):
             obj['result_ttl'] = self.result_ttl
         if self._status is not None:
             obj['status'] = self._status
+        if self._dependency_id is not None:
+            obj['dependency_id'] = self._dependency_id
         if self.meta:
             obj['meta'] = dumps(self.meta)
 
-        connection.hmset(key, obj)
+        return obj
+
+    def save(self, pipeline=None):
+        """Persists the current job instance to its corresponding Redis key."""
+        key = self.key
+        connection = pipeline if pipeline is not None else self.connection
+
+        connection.hmset(key, self.dump())
 
     def cancel(self):
         """Cancels the given job, which will prevent the job from ever being
@@ -333,23 +433,29 @@ class Job(object):
         cancellation.  Technically, this call is (currently) the same as just
         deleting the job hash.
         """
-        self.delete()
+        pipeline = self.connection._pipeline()
+        self.delete(pipeline=pipeline)
+        pipeline.delete(self.dependents_key)
+        pipeline.execute()
 
-    def delete(self):
+    def delete(self, pipeline=None):
         """Deletes the job hash from Redis."""
-        self.connection.delete(self.key)
-
+        connection = pipeline if pipeline is not None else self.connection
+        connection.delete(self.key)
 
     # Job execution
     def perform(self):  # noqa
         """Invokes the job function with the job arguments."""
         _job_stack.push(self.id)
         try:
+            self.set_status(Status.STARTED)
             self._result = self.func(*self.args, **self.kwargs)
+            self.set_status(Status.FINISHED)
+            self.ended_at = utcnow()
         finally:
             assert self.id == _job_stack.pop()
-        return self._result
 
+        return self._result
 
     def get_ttl(self, default_ttl=None):
         """Returns ttl for a job that determines how long a job and its result
@@ -379,18 +485,29 @@ class Job(object):
         - If it's a positive number, set the job to expire in X seconds.
         - If result_ttl is negative, don't set an expiry to it (persist
           forever)
-        """        
+        """
         if ttl == 0:
             self.cancel()
         elif ttl > 0:
             connection = pipeline if pipeline is not None else self.connection
             connection.expire(self.key, ttl)
-        
+
+    def register_dependency(self):
+        """Jobs may have dependencies. Jobs are enqueued only if the job they
+        depend on is successfully performed. We record this relation as
+        a reverse dependency (a Redis set), with a key that looks something
+        like:
+
+            rq:job:job_id:dependents = {'job_id_1', 'job_id_2'}
+
+        This method adds the current job in its dependency's dependents set.
+        """
+        # TODO: This can probably be pipelined
+        self.connection.sadd(Job.dependents_key_for(self._dependency_id), self.id)
 
     def __str__(self):
         return '<Job %s: %s>' % (self.id, self.description)
 
-
     # Job equality
     def __eq__(self, other):  # noqa
         return self.id == other.id
@@ -398,5 +515,4 @@ class Job(object):
     def __hash__(self):
         return hash(self.id)
 
-
 _job_stack = LocalStack()
