@@ -3,7 +3,6 @@ import os
 import errno
 import random
 import time
-import times
 try:
     from procname import setprocname
 except ImportError:
@@ -16,10 +15,10 @@ import logging
 from .queue import Queue, get_failed_queue
 from .connections import get_current_connection
 from .job import Job, Status
-from .utils import make_colorizer
+from .utils import make_colorizer, utcnow, utcformat
 from .logutils import setup_loghandlers
-from .exceptions import NoQueueError, UnpickleError, DequeueTimeout
-from .timeouts import death_penalty_after
+from .exceptions import NoQueueError, DequeueTimeout
+from .timeouts import UnixSignalDeathPenalty
 from .version import VERSION
 from rq.compat import text_type, as_text
 
@@ -43,9 +42,9 @@ def iterable(x):
 def compact(l):
     return [x for x in l if x is not None]
 
-_signames = dict((getattr(signal, signame), signame) \
-                    for signame in dir(signal) \
-                    if signame.startswith('SIG') and '_' not in signame)
+_signames = dict((getattr(signal, signame), signame)
+                 for signame in dir(signal)
+                 if signame.startswith('SIG') and '_' not in signame)
 
 
 def signal_name(signum):
@@ -60,6 +59,8 @@ def signal_name(signum):
 class Worker(object):
     redis_worker_namespace_prefix = 'rq:worker:'
     redis_workers_keys = 'rq:workers'
+    death_penalty_class = UnixSignalDeathPenalty
+
 
     @classmethod
     def all(cls, connection=None):
@@ -68,8 +69,8 @@ class Worker(object):
         if connection is None:
             connection = get_current_connection()
         reported_working = connection.smembers(cls.redis_workers_keys)
-        workers = [cls.find_by_key(as_text(key), connection) for key in
-                reported_working]
+        workers = [cls.find_by_key(as_text(key), connection)
+                   for key in reported_working]
         return compact(workers)
 
     @classmethod
@@ -79,7 +80,6 @@ class Worker(object):
         by their Redis keys.
         """
         prefix = cls.redis_worker_namespace_prefix
-        name = worker_key[len(prefix):]
         if not worker_key.startswith(prefix):
             raise ValueError('Not a valid RQ worker key: %s' % (worker_key,))
 
@@ -93,15 +93,15 @@ class Worker(object):
         worker = cls([], name, connection=connection)
         queues = as_text(connection.hget(worker.key, 'queues'))
         worker._state = connection.hget(worker.key, 'state') or '?'
+        worker._job_id = connection.hget(worker.key, 'current_job') or None
         if queues:
             worker.queues = [Queue(queue, connection=connection)
-                                for queue in queues.split(',')]
+                             for queue in queues.split(',')]
         return worker
 
-
     def __init__(self, queues, name=None,
-            default_result_ttl=DEFAULT_RESULT_TTL, connection=None,
-            exc_handler=None, default_worker_ttl=DEFAULT_WORKER_TTL):  # noqa
+                 default_result_ttl=None, connection=None,
+                 exc_handler=None, default_worker_ttl=None):  # noqa
         if connection is None:
             connection = get_current_connection()
         self.connection = connection
@@ -111,8 +111,15 @@ class Worker(object):
         self.queues = queues
         self.validate_queues()
         self._exc_handlers = []
+
+        if default_result_ttl is None:
+            default_result_ttl = DEFAULT_RESULT_TTL
         self.default_result_ttl = default_result_ttl
+
+        if default_worker_ttl is None:
+            default_worker_ttl = DEFAULT_WORKER_TTL
         self.default_worker_ttl = default_worker_ttl
+
         self._state = 'starting'
         self._is_horse = False
         self._horse_pid = 0
@@ -193,15 +200,13 @@ class Worker(object):
         self.log.debug('Registering birth of worker %s' % (self.name,))
         if self.connection.exists(self.key) and \
                 not self.connection.hexists(self.key, 'death'):
-            raise ValueError(
-                    'There exists an active worker named \'%s\' '
-                    'already.' % (self.name,))
+            raise ValueError('There exists an active worker named \'%s\' '
+                             'already.' % (self.name,))
         key = self.key
-        now = time.time()
         queues = ','.join(self.queue_names())
         with self.connection._pipeline() as p:
             p.delete(key)
-            p.hset(key, 'birth', now)
+            p.hset(key, 'birth', utcformat(utcnow()))
             p.hset(key, 'queues', queues)
             p.sadd(self.redis_workers_keys, key)
             p.expire(key, self.default_worker_ttl)
@@ -214,18 +219,54 @@ class Worker(object):
             # We cannot use self.state = 'dead' here, because that would
             # rollback the pipeline
             p.srem(self.redis_workers_keys, self.key)
-            p.hset(self.key, 'death', time.time())
+            p.hset(self.key, 'death', utcformat(utcnow()))
             p.expire(self.key, 60)
             p.execute()
 
-    def set_state(self, new_state):
-        self._state = new_state
-        self.connection.hset(self.key, 'state', new_state)
+    def set_state(self, state, pipeline=None):
+        self._state = state
+        connection = pipeline if pipeline is not None else self.connection
+        connection.hset(self.key, 'state', state)
+
+    def _set_state(self, state):
+        """Raise a DeprecationWarning if ``worker.state = X`` is used"""
+        raise DeprecationWarning(
+            "worker.state is deprecated, use worker.set_state() instead."
+        )
+        self.set_state(state)
 
     def get_state(self):
         return self._state
 
-    state = property(get_state, set_state)
+    def _get_state(self):
+        """Raise a DeprecationWarning if ``worker.state == X`` is used"""
+        raise DeprecationWarning(
+            "worker.state is deprecated, use worker.get_state() instead."
+        )
+        return self.get_state()
+
+    state = property(_get_state, _set_state)
+
+    def set_current_job_id(self, job_id, pipeline=None):
+        connection = pipeline if pipeline is not None else self.connection
+
+        if job_id is None:
+            connection.hdel(self.key, 'current_job')
+        else:
+            connection.hset(self.key, 'current_job', job_id)
+
+    def get_current_job_id(self, pipeline=None):
+        connection = pipeline if pipeline is not None else self.connection
+        return as_text(connection.hget(self.key, 'current_job'))
+
+    def get_current_job(self):
+        """Returns the job id of the currently executing job."""
+        job_id = self.get_current_job_id()
+
+        if job_id is None:
+            return None
+
+        return Job.fetch(job_id, self.connection)
 
     @property
     def stopped(self):
@@ -268,7 +309,7 @@ class Worker(object):
 
             # If shutdown is requested in the middle of a job, wait until
             # finish before shutting down
-            if self.state == 'busy':
+            if self.get_state() == 'busy':
                 self._stopped = True
                 self.log.debug('Stopping after current horse is finished. '
                                'Press Ctrl+C again for a cold shutdown.')
@@ -294,18 +335,13 @@ class Worker(object):
         did_perform_work = False
         self.register_birth()
         self.log.info('RQ worker started, version %s' % VERSION)
-        self.state = 'starting'
+        self.set_state('starting')
         try:
             while True:
                 if self.stopped:
                     self.log.info('Stopping on request.')
                     break
-                self.state = 'idle'
-                qnames = self.queue_names()
-                self.procline('Listening on %s' % ','.join(qnames))
-                self.log.info('')
-                self.log.info('*** Listening on %s...' % \
-                        green(', '.join(qnames)))
+
                 timeout = None if burst else max(1, self.default_worker_ttl - 60)
                 try:
                     result = self.dequeue_job_and_maintain_ttl(timeout)
@@ -313,37 +349,41 @@ class Worker(object):
                         break
                 except StopRequested:
                     break
-                except UnpickleError as e:
-                    job = Job.safe_fetch(e.job_id)
-                    self.handle_exception(job, *sys.exc_info())
-                    continue
-
-                self.state = 'busy'
 
                 job, queue = result
-                # Use the public setter here, to immediately update Redis
-                job.status = Status.STARTED
-                self.log.info('%s: %s (%s)' % (green(queue.name),
-                    blue(job.description), job.id))
-
-                self.heartbeat((job.timeout or 180) + 60)
-                self.fork_and_perform_job(job)
+                self.execute_job(job)
                 self.heartbeat()
 
+                if job.get_status() == Status.FINISHED:
+                    queue.enqueue_dependents(job)
+
                 did_perform_work = True
         finally:
             if not self.is_horse:
                 self.register_death()
         return did_perform_work
 
-
     def dequeue_job_and_maintain_ttl(self, timeout):
         result = None
+        qnames = self.queue_names()
+
+        self.set_state('idle')
+        self.procline('Listening on %s' % ','.join(qnames))
+        self.log.info('')
+        self.log.info('*** Listening on %s...' %
+                      green(', '.join(qnames)))
+
         while True:
             self.heartbeat()
+
             try:
                 result = Queue.dequeue_any(self.queues, timeout,
                                            connection=self.connection)
+                if result is not None:
+                    job, queue = result
+                    self.log.info('%s: %s (%s)' % (green(queue.name),
+                                  blue(job.description), job.id))
+
                 break
             except DequeueTimeout:
                 pass
@@ -367,7 +407,7 @@ class Worker(object):
         self.log.debug('Sent heartbeat to prevent worker timeout. '
                        'Next one should arrive within {0} seconds.'.format(timeout))
 
-    def fork_and_perform_job(self, job):
+    def execute_job(self, job):
         """Spawns a work horse to perform the actual work and passes it a job.
         The worker will wait for the work horse and make sure it executes
         within the given timeout bounds, or will end the work horse with
@@ -420,32 +460,38 @@ class Worker(object):
         """Performs the actual work of a job.  Will/should only be called
         inside the work horse's process.
         """
+
+        self.set_state('busy')
+        self.set_current_job_id(job.id)
+        self.heartbeat((job.timeout or 180) + 60)
+
         self.procline('Processing %s from %s since %s' % (
             job.func_name,
             job.origin, time.time()))
 
-        try:
-            with death_penalty_after(job.timeout or 180):
-                rv = job.perform()
+        with self.connection._pipeline() as pipeline:
+            try:
+                with self.death_penalty_class(job.timeout or Queue.DEFAULT_TIMEOUT):
+                    rv = job.perform()
 
-            # Pickle the result in the same try-except block since we need to
-            # use the same exc handling when pickling fails
-            job._result = rv
-            job._status = Status.FINISHED
-            job.ended_at = times.now()
+                # Pickle the result in the same try-except block since we need to
+                # use the same exc handling when pickling fails
+                job._result = rv
 
-            result_ttl = job.get_ttl(self.default_result_ttl)
-            pipeline = self.connection._pipeline()
-            if result_ttl != 0:
-                job.save(pipeline=pipeline)
-            job.cleanup(result_ttl, pipeline=pipeline)
-            pipeline.execute()
+                self.set_current_job_id(None, pipeline=pipeline)
 
-        except:
-            # Use the public setter here, to immediately update Redis
-            job.status = Status.FAILED
-            self.handle_exception(job, *sys.exc_info())
-            return False
+                result_ttl = job.get_ttl(self.default_result_ttl)
+                if result_ttl != 0:
+                    job.save(pipeline=pipeline)
+                job.cleanup(result_ttl, pipeline=pipeline)
+
+                pipeline.execute()
+
+            except Exception:
+                # Use the public setter here, to immediately update Redis
+                job.set_status(Status.FAILED)
+                self.handle_exception(job, *sys.exc_info())
+                return False
 
         if rv is None:
             self.log.info('Job OK')
@@ -461,12 +507,10 @@ class Worker(object):
 
         return True
 
-
     def handle_exception(self, job, *exc_info):
         """Walks the exception handler stack to delegate exception handling."""
-        exc_string = ''.join(
-                traceback.format_exception_only(*exc_info[:2]) +
-                traceback.format_exception(*exc_info))
+        exc_string = ''.join(traceback.format_exception_only(*exc_info[:2]) +
+                             traceback.format_exception(*exc_info))
         self.log.error(exc_string)
 
         for handler in reversed(self._exc_handlers):
